#include "Essentials.h"
#include "Filing.h"
#include "Ready Menus.h"

static void		ClearDisposeMenu(MenuRef menu);
#define				hierarchicals	10
#define				baseMenuID		256

static void			BuildSubMenu(int hierarchical, FSSpec* fileSpec);
pascal OSStatus		CarbonEventSubMenu(EventHandlerCallRef callRef, EventRef event, void *data);

pascal void			FolderChangeNotification(FNMessage message, OptionBits flags, void *refcon, FNSubscriptionRef subscription);
FNSubscriptionUPP	FolderChangeNotificationUPP;

static MenuRef		subMenuRef[hierarchicals];
static ItemCount	subMenuItems[hierarchicals];

void InitializeFolderCachedMenu(ReadyMenu* readyMenu)
{
	MenuRef		saveSubMenuRef; // coz we can call this in middle of ClickFolderMenu
	FSRef		fileRef;
	
	readyMenu->menu=GetMenu(161);
	readyMenu->notifyRef=0;
	saveSubMenuRef=subMenuRef[0];
	subMenuRef[0]=readyMenu->menu;
	subMenuItems[0]=0;
	BuildSubMenu(0, &readyMenu->spec->file);
	subMenuRef[0]=saveSubMenuRef;
	
	if (!FolderChangeNotificationUPP)
		FolderChangeNotificationUPP=NewFNSubscriptionUPP(FolderChangeNotification);
	
	if (FolderChangeNotificationUPP)
		if (FSpMakeFSRef(&readyMenu->spec->file, &fileRef)==0)
			if (FNSubscribe(&fileRef, FolderChangeNotificationUPP, readyMenu,
					(1 << 1) /*kFNNotifyInBackground - not in 10.2.8 SDK but is on by default in earlier OS X versions*/, &readyMenu->notifyRef))
				readyMenu->notifyRef=0;
}

void DestroyFolderCachedMenu(ReadyMenu* readyMenu)
{
	if (readyMenu) {		
		if (readyMenu->menu) {
			if (readyMenu->notifyRef)
				FNUnsubscribe(readyMenu->notifyRef);
			
			DisposeMenu(readyMenu->menu);
			
			readyMenu->menu=NULL;
		}
	
	} else {
		for (readyMenu=firstReadyMenu; readyMenu; readyMenu=readyMenu->next)
			DestroyFolderCachedMenu(readyMenu);
	}
}

pascal void FolderChangeNotification(FNMessage message, OptionBits flags, void *refcon, FNSubscriptionRef subscription)
{
	ReadyMenu	*readyMenu;
	
	if (message==kFNDirectoryModifiedMessage) {
		readyMenu=(ReadyMenu*)refcon;
		DestroyFolderCachedMenu(readyMenu);
	}
}

void ClickFolderMenu(Point globalPoint, ReadyMenu* readyMenu, DragReference dragRef)
{
    int					hierarchical;
    long				selection;
    FSSpec				openFileSpec;
    UInt32				actualSize;
    EventHandlerUPP		CarbonEventSubMenuUPP;
    EventHandlerRef		CarbonEventSubMenuRef[hierarchicals];
    EventTargetRef		targetRef;
    EventTypeSpec		eventList[]={
							{kEventClassMenu, kEventMenuOpening},
						};
    short				menuID, menuItem, menuItems;
	ProcessSerialNumber	process;
	Boolean				isPathFinder;
	UInt32				modifiers;
	FinderAction		action;
	
	if (readyMenu->menu==NULL)
		InitializeFolderCachedMenu(readyMenu);
	
	CarbonEventSubMenuUPP=NewEventHandlerUPP(CarbonEventSubMenu);
    
    for (hierarchical=0; hierarchical<hierarchicals; hierarchical++) {
        if (hierarchical==0) {
			subMenuRef[hierarchical]=readyMenu->menu;				
            SetMenuID(subMenuRef[hierarchical], baseMenuID+hierarchical);
            menuItems=CountMenuItems(subMenuRef[hierarchical]);
        } else
            CreateNewMenu(baseMenuID+hierarchical, 0, &subMenuRef[hierarchical]);
            
        InsertMenu(subMenuRef[hierarchical], -1);
        subMenuItems[hierarchical]=0;

		targetRef=GetMenuEventTarget(subMenuRef[hierarchical]);
        InstallEventHandler(targetRef, CarbonEventSubMenuUPP,
            sizeof(eventList)/sizeof(EventTypeSpec), eventList, 0, &CarbonEventSubMenuRef[hierarchical]);
    }
    
	selection=PopUpMenuSelect(subMenuRef[0], globalPoint.v, globalPoint.h, 0);
	if (selection==0)
        selection=MenuChoice();

	modifiers=GetCurrentKeyModifiers();
	
	for (hierarchical=0; hierarchical<hierarchicals; hierarchical++)
		DeleteMenu(GetMenuID(subMenuRef[hierarchical])); // take out of menu list

	menuID=HiWord(selection);
    menuItem=LoWord(selection);
    
    if ((menuID>=baseMenuID)&&(menuID<(baseMenuID+hierarchicals))) {
		if (dragRef) {
		
		} else {
			if (GetMenuItemProperty(subMenuRef[menuID-baseMenuID], menuItem,
				'ClaM', 'fss ', sizeof(FSSpec), &actualSize, &openFileSpec)==0) {
				
				if (modifiers&cmdKey)
					action=finderReveal;
				else if (modifiers&shiftKey)
					action=finderGetInfo;
				else
					action=finderOpen;
				
				// IncrementSelectionCount();
				_i(DoFinderFile(action, &openFileSpec, NULL));
			}
		}
                
    } else if (menuID==224)
        SelectClassicMenuItem(menuItem);
        
    else if (menuID==225) {
        switch (menuItem) {
            case 1: // add alias
                if (GetFinderProcess(&process, &isPathFinder)==0) {
					if (isPathFinder)
						goto openInFinder; // can't create alias via PF

					else if (SelectAnObject(&openFileSpec)) {
						CreateAlias(&openFileSpec, &readyMenu->spec->file);
						DestroyFolderCachedMenu(readyMenu);
					}
				}
                break;
                
            case 2: openInFinder: // open in finder
                BringFinderToFront();
                DoFinderFile(finderOpen, &readyMenu->spec->file, NULL);
                break;
            
            case 3: // select new
                if (SelectNewFolder(&readyMenu->spec->file))
					DestroyFolderCachedMenu(readyMenu);
				break;

        }
    }
    
    for (hierarchical=0; hierarchical<hierarchicals; hierarchical++) {
        RemoveEventHandler(CarbonEventSubMenuRef[hierarchical]);
        
		if (hierarchical>0) // don't kill menu or free icons
			ClearDisposeMenu(subMenuRef[hierarchical]);
    }
	
    DisposeEventHandlerUPP(CarbonEventSubMenuUPP);
    
    _e
}

void SelectClassicMenuItem(short item)
{
    switch (item) {
        case 1: // about
            // ShowAboutDialog();
            break;
            
        case 2: // preferences
            ShowPreferencesDialog();
			break;
        
        case 3: // register
            // ShowRegisterDialog();
            break;
            
        case 4: // quit
            QuitApplicationEventLoop();
            break;
    }
}

pascal OSStatus CarbonEventSubMenu(EventHandlerCallRef callRef, EventRef event, void *data)
{
    OSStatus			error;
    UInt32				class, kind;
    EventParamType		actualType;
    UInt32				actualSize;
    MenuRef				menu;
    int					hierarchical;
    MenuID				menuID;
	MenuTrackingData	trackingData;
    FSSpec				buildFileSpec;
	
    error=GetEventParameter(event, kEventParamDirectObject, typeMenuRef, &actualType,
            sizeof(MenuRef), &actualSize, &menu);
        _i(error)
        
    menuID=GetMenuID(menu);
    hierarchical=menuID-baseMenuID;

    class=GetEventClass(event);
    kind=GetEventKind(event);
    	
	if ((class==kEventClassMenu)&&(kind==kEventMenuOpening)) {
		if (hierarchical>0) {
			GetMenuTrackingData(subMenuRef[hierarchical-1], &trackingData);
			GetMenuItemProperty(trackingData.menu, trackingData.itemSelected, 'ClaM', 'fss ',
				sizeof(FSSpec), &actualSize, &buildFileSpec); // 2.5.2 much better I hope than relying on kEventMenuTargetItem event
			BuildSubMenu(hierarchical, &buildFileSpec);
		}

	}

    _e
    return error;
}

static void BuildSubMenu(int hierarchical, FSSpec* fileSpec)
{	
    Error				error;
    FSRef				fileRef, testRef;
    MenuRef				menu;
    FSCatalogInfo		catalogInfo, *catalogInfoArray;
    FSIterator			iterator;
    FSSpec				*fileSpecArray, *fileSpecEntry;
    FSRef				*fileRefArray;
	UInt32				fileIndex, testFileIndex;
    IconRef				iconRef;
    SInt16				theLabel;
    Boolean				isAlias, isFolder, showItemIcon, menuIsVolumeRoot;
    Str255				emptyString;
	short				fileRefNum;
	unsigned char		*hiddenFile, *hiddenFilePtr, *hiddenFileEnd;
	long				hiddenFileSize;
	MenuItemIndex		insertPosition, minInsertPosition, maxInsertPosition;
	CFStringRef			cfString, compareString;
	CFRange				range;
	CFComparisonResult	result;
	LSItemInfoRecord	testInfo;
	unsigned char		*rootHideFiles[]={ "\pmach", "\pmach.sym" };
	int					hideIndex;
	
    menu=subMenuRef[hierarchical];
    DeleteMenuItems(menu, 1, subMenuItems[hierarchical]);
    subMenuItems[hierarchical]=0;
    
    error=FSpMakeFSRef(fileSpec, &fileRef);
        _i(error)
    
	error=FSGetCatalogInfo(&fileRef, kFSCatInfoValence|kFSCatInfoParentDirID, &catalogInfo, NULL, NULL, NULL);
    	_i(error)
		
	menuIsVolumeRoot=(catalogInfo.parentDirID==fsRtParID);
    
    fileSpecArray=(FSSpec*)AllocPtr(&error, sizeof(FSSpec)*catalogInfo.valence);
        _i(error)
    
    catalogInfoArray=(FSCatalogInfo*)AllocPtr(&error, sizeof(FSCatalogInfo)*catalogInfo.valence);
        _i2(error)
    
	fileRefArray=(FSRef*)AllocPtr(&error, sizeof(FSRef)*catalogInfo.valence);
		_i3(error)
	
    error=FSOpenIterator(&fileRef, kFSIterateFlat, &iterator);
        _i4(error)
    
    error=FSGetCatalogInfoBulk(iterator, catalogInfo.valence, &catalogInfo.valence, NULL,
        kFSCatInfoFinderInfo, catalogInfoArray, fileRefArray, fileSpecArray, NULL);
    if (error==errFSNoMoreItems)
        error=0;
        _i5(error)
	
	// special code to remove mach and mach.sym from OS X 10.4.x, since .hidden file is gone now
	// In 10.4.x, other files in similar situation are marked as hidden in their CatalogInfo
	
	if (menuIsVolumeRoot)
		for (fileIndex=0; fileIndex<catalogInfo.valence; fileIndex++) {
			for (hideIndex=0; hideIndex<(sizeof(rootHideFiles)/sizeof(*rootHideFiles)); hideIndex++)
				if (EqualString(fileSpecArray[fileIndex].name, rootHideFiles[hideIndex], true, true)) {
					((FileInfo*)(catalogInfoArray[fileIndex].finderInfo))->finderFlags|=kIsInvisible;
					goto rootFileHidden;
				}
			
			rootFileHidden:;
		}
			
    for (fileIndex=0; fileIndex<catalogInfo.valence; fileIndex++)
		if (EqualString(fileSpecArray[fileIndex].name, "\p.hidden", true, true)) {
			error=FSpOpenDF(&fileSpecArray[fileIndex], fsRdPerm, &fileRefNum);
				_i6(error)
				
			error=GetEOF(fileRefNum, &hiddenFileSize);
				_i7(error)
				
			hiddenFile=(unsigned char*)AllocPtr(&error, hiddenFileSize);
				_i7(error)
				
			error=FSRead(fileRefNum, &hiddenFileSize, hiddenFile);
				_i8(error)
				
			hiddenFilePtr=hiddenFile;
			hiddenFileEnd=hiddenFilePtr+hiddenFileSize;
			*emptyString=0;
			
			while (hiddenFilePtr<hiddenFileEnd) {
				switch (*hiddenFilePtr) {
					case 0x0A: case 0x0D:
						for (testFileIndex=0; testFileIndex<catalogInfo.valence; testFileIndex++)
							if (EqualString(fileSpecArray[testFileIndex].name, emptyString, true, true))
								((FileInfo*)(catalogInfoArray[testFileIndex].finderInfo))->finderFlags|=kIsInvisible;
						*emptyString=0;
						break;
						
					default:
						emptyString[++*emptyString]=*hiddenFilePtr;
						break;
				}
				
				hiddenFilePtr++;
			}
				
			_e8
			DestroyPtr(hiddenFile);
			_e7
			FSClose(fileRefNum);
			_e6
		}

	for (fileIndex=0; fileIndex<catalogInfo.valence; fileIndex++) {
        if ((fileSpecArray[fileIndex].name[1]!='.')&&(!(((FileInfo*)(catalogInfoArray[fileIndex].finderInfo))->finderFlags&kIsInvisible))) {
            fileSpecEntry=&fileSpecArray[fileIndex];
        
		//	Build CFString to add
			
			LSCopyDisplayNameForRef(&fileRefArray[fileIndex], &cfString);
			
		//	Find where to insert it
			
			insertPosition=subMenuItems[hierarchical]+1;
			
			if (CopyMenuItemTextAsCFString(menu, subMenuItems[hierarchical], &compareString)==0) {
				result=CFStringCompare(cfString, compareString, kCFCompareCaseInsensitive|kCFCompareNonliteral);
				CFRelease(compareString);
				
				if (result<kCFCompareEqualTo) {
					minInsertPosition=1;
					maxInsertPosition=subMenuItems[hierarchical];
					
					while (minInsertPosition<maxInsertPosition) {
						insertPosition=(minInsertPosition+maxInsertPosition)/2;
						
						if (CopyMenuItemTextAsCFString(menu, insertPosition, &compareString)==0) {
							result=CFStringCompare(cfString, compareString, kCFCompareCaseInsensitive|kCFCompareNonliteral);
							CFRelease(compareString);
							
							if (result<kCFCompareEqualTo)
								maxInsertPosition=insertPosition;
							else
								minInsertPosition=insertPosition+1;
						}
					}
					
					insertPosition=maxInsertPosition;
				}
			}
		
		//	Insert and release
			
			InsertMenuItemTextWithCFString(menu, cfString, insertPosition-1, 0, 0);
			ChangeMenuItemAttributes(menu, insertPosition, 0, kMenuItemAttrSeparator); // make sure not a separator
			CFRelease(cfString);
			
			subMenuItems[hierarchical]++; // to make sure it's always correct even if something fails
            
        //	Decide whether to show icon and which to show
		
			showItemIcon=(hierarchical>0)
				? (generalSettings.showIconsSubmenus&&(catalogInfo.valence<=generalSettings.submenuIconLimit))
				: generalSettings.showIconsMainMenu;
				
		//	Generate hierarchy child if necessary (also resolves alias)
			
			if ((hierarchical+1)<hierarchicals)
				if (ResolveAliasFileWithMountFlags(fileSpecEntry, true, &isFolder, &isAlias, kResolveAliasFileNoUI)==0)
                    if (isFolder)
						if (FSpMakeFSRef(fileSpecEntry, &testRef)==0)
							if (LSCopyItemInfoForRef(&testRef, kLSRequestBasicFlagsOnly, &testInfo)==0)
								if ((testInfo.flags&kLSItemInfoIsContainer)&&!(testInfo.flags&kLSItemInfoIsPackage)) {
									SetMenuItemHierarchicalID(menu, insertPosition, baseMenuID+hierarchical+1);
									ChangeMenuItemAttributes(menu, insertPosition, kMenuItemAttrSubmenuParentChoosable, 0);
								}

        //	Add icon after resolving alias to prevent arrow showing for others
				
            if (showItemIcon/*&&!showCustomIcon*/)
				if (GetIconRefFromFile(fileSpecEntry, &iconRef, &theLabel)==0)
					SetMenuItemIconHandle(menu, insertPosition, kMenuIconRefType, (Handle)iconRef);
            
		//	Add file reference to menu item

			SetMenuItemProperty(menu, insertPosition, 'ClaM', 'fss ', sizeof(FSSpec), fileSpecEntry);
        }
    }
	
//	Now post-process the menu items to remove any >> prefixes (these were used as part of ordering)
//	Also now remove any prefixes where the 3rd character is a ')' and deal with FinderPop-style separators

	for (insertPosition=1; insertPosition<=subMenuItems[hierarchical]; insertPosition++)
		if (CopyMenuItemTextAsCFString(menu, insertPosition, &cfString)==0) {
			range=CFStringFind(cfString, CFSTR(">>"), 0);
			
			if (range.location>=0) {
				compareString=CFStringCreateMutableCopy(NULL, 0, cfString);
				range.length=range.location+2;
				range.location=0;
				CFStringDelete(compareString, range);
				SetMenuItemTextWithCFString(menu, insertPosition, compareString);
				
				if (CFStringCompare(compareString, CFSTR("-"), 0)==kCFCompareEqualTo)
					ChangeMenuItemAttributes(menu, insertPosition, kMenuItemAttrSeparator, 0); // set to a separator
				
				CFRelease(compareString);
			
			} else {
				range=CFStringFind(cfString, CFSTR(")"), 0);
				
				if (range.location==2) {
					compareString=CFStringCreateMutableCopy(NULL, 0, cfString);
					range.length=range.location+1;
					range.location=0;
					CFStringDelete(compareString, range);
					SetMenuItemTextWithCFString(menu, insertPosition, compareString);
					
					if (CFStringCompare(compareString, CFSTR("-***"), 0)==kCFCompareEqualTo)
						ChangeMenuItemAttributes(menu, insertPosition, kMenuItemAttrSeparator, 0); // set to a separator
					
					CFRelease(compareString);
				}
			}
			
			CFRelease(cfString);
		}

	_e5
    FSCloseIterator(iterator);
    _e4
	DestroyPtr(fileRefArray);
	_e3
    DestroyPtr(catalogInfoArray);
    _e2
    DestroyPtr(fileSpecArray);
    _e

    if (subMenuItems[hierarchical]==0) {
        GetIndString(emptyString, 128, 1);
        InsertMenuItemText(menu, emptyString, subMenuItems[hierarchical]);
        subMenuItems[hierarchical]++;
    }
}

void ClearDisposeMenu(MenuRef menu)
{
    SInt16		menuItem, menuItems;
    UInt8		iconType;
    Handle		iconHandle;
    
    menuItems=CountMenuItems(menu);
    for (menuItem=1; menuItem<menuItems; menuItem++) {
        if (GetMenuItemIconHandle(menu, menuItem, &iconType, &iconHandle)==0)
            if (iconType==kMenuIconRefType)
                ReleaseIconRef((IconRef)iconHandle);
    }
     
    DisposeMenu(menu);
}