#include <Carbon/Carbon.h>

#include "User Settings.h"

typedef 	OSErr Error;

void		TestError(Error error);
Error		TestResError(void* resource);
Error		TestMemError(void* pointer);

Boolean		HideErrors(Error* keepError);
void		ShowErrors();

Boolean		AddErrorContext(Str255 contextString);
void		RemoveErrorContext();

#if Project_Under_Development

void		HandleAssertFailure(char* file, char* date, int line); // params shown in dialog

#define		Assert(test)	{ if (!(test)) HandleAssertFailure(__FILE__, __DATE__, __LINE__); }

#else

#define		Assert(test)	;

#endif

#define		AssertRange(test, minimum, maximum)	Assert(((test)>=(minimum))&&((test)<=(maximum)));

#if Project_Under_Development

void		RegisterPointer(void* pointer, OSType signature);
void		UnregisterPointer(void* pointer, OSType signature);
Boolean		PointerExists(void* pointer, OSType signature);
				
#else

#define		RegisterPointer(p, s)	;
#define		UnregisterPointer(p, s)	;
#define		PointerExists(p, s)		true

#endif

void		AddToLinkedList(void** firstPointer, void* addItem, void* beforeItem);
void		RemoveFromLinkedList(void** firstPointer, void* removeItem);
void*		GetLinkedListItem(void* firstPointer, long itemNumber);
long		CountLinkedListItems(void* firstPointer);

void*		AllocPtr(Error* error, Size allocate);
void**		AllocHandle(Error* error, Size allocate);
void		DestroyPtr(void* pointer);
void		DestroyHandle(void** handle);

#if Project_Under_Development && Assert_Memory_Locking

void		LockHandleAssert(void** handle);
void		UnlockHandleAssert(void** handle);

#define		LockHandle(handle) LockHandleAssert(handle)
#define 	UnlockHandle(handle) UnlockHandleAssert(handle)

#else

#define		LockHandle(handle) HLock((Handle)(handle))
#define		UnlockHandle(handle) HUnlock((Handle)(handle))

#endif

#define _e		failed:;
#define	_e2		failed2:;
#define _e3		failed3:;
#define _e4		failed4:;
#define _e5		failed5:;
#define	_e6		failed6:;
#define	_e7		failed7:;
#define	_e8		failed8:;
#define	_e9		failed9:;

#define _g		goto failed;
#define _g2		goto failed2;
#define _g3		goto failed3;
#define _g4		goto failed4;
#define _g5		goto failed5;
#define	_g6		goto failed6;
#define	_g7		goto failed7;
#define	_g8		goto failed8;
#define	_g9		goto failed9;

#define	_i(p)	{ if (p) _g }
#define	_i2(p)	{ if (p) _g2 }
#define	_i3(p)	{ if (p) _g3 }
#define	_i4(p)	{ if (p) _g4 }
#define	_i5(p)	{ if (p) _g5 }
#define	_i6(p)	{ if (p) _g6 }
#define	_i7(p)	{ if (p) _g7 }
#define	_i8(p)	{ if (p) _g8 }
#define	_i9(p)	{ if (p) _g9 }

void			GlueSetTheCursorToArrow();
Size			GlueGetScrap(Handle handle, OSType type);

void			SetItemHandle(DialogPtr dialog, short item, Handle handle);
Handle			GetItemHandle(DialogPtr dialog, short item);
void			GetItemRect(DialogRef dialog, short item, Rect* rect);
void			SetItemValue(DialogPtr dialog, short item, short value);
short			GetItemValue(DialogRef dialog, short item);
void			SetItemEnabled(DialogPtr dialog, short item, Boolean enabled);
Boolean			ToggleItem(DialogRef dialog, short item);
void			GetItemText(DialogPtr dialog, short item, Str255 string);
long			GetItemNumber(DialogPtr dialog, short item);
void			SetItemText(DialogPtr dialog, short item, const Str255 string);
void			SetItemNumber(DialogPtr dialog, short item, long number);

Error			FindApplication(OSType creator, ProcessSerialNumber *process);
Error			FindApplicationProcess(OSType creator, ProcessSerialNumber *process);
Error			FindApplicationFile(OSType creator, FSSpec* fileSpec);
Error			OpenApplicationFile(FSSpec* fileSpec, ProcessSerialNumber *process);

#define	memoryBufferBlockSize	1024
#define	saveBufferBlockSize		1024
#define	loadBufferBlockSize		1024

typedef struct MemoryBuffer {
	Handle		handle;
	long		allocated, used;
} MemoryBuffer;

Error			CreateMemoryBuffer(MemoryBuffer* buffer);
Error			WriteMemoryBuffer(MemoryBuffer* buffer, const void* dataPtr, long count);
Handle			CloseMemoryBuffer(MemoryBuffer* buffer);
