#include "Essentials.h"
#include "Ready Menus.h"
#include "Filing.h"
#include "LoginItemAPI.h"
#include "Global PowerPC Intel Glue.h"

enum {
    prefsEndMarker='end!',
    prefsRegistration='regi',
	prefsAppleSettings='apse',
	prefsGeneralSettings='gese',
    prefsMenuSpec='mspc',
	prefsMenuAlias='mals', // always for the menuSpec that just came before
    prefsSelectionCount='selc'
};

static Error		LoadPreference(short fileRefNum, long sizeInFile, void* loadPtr, long sizeInMemory);
static Error		SavePreference(short fileRefNum, OSType signature, void* savePtr, long saveSize);

static Error		GetPrefsFile(FSSpec* file);
static Error		SetDefaultPreferences();

MenuSpec			*firstMenuSpec=0;
unsigned long		selectionCount=0;
Registration		registration={"\p", 0, 0};
short				registered=0;
Boolean				registrationToUpdate;
AppleMenuSettings	appleMenuSettings={
						colorRainbow,
						showSystemWhenModifierDown,
						controlKey,
						mouseWithinLeft
					};
GeneralSettings		generalSettings={
						true,
						true,
						true,
						true,
						50
					};

void ToLoadMenuSpec(MenuSpec* spec)
{
	ToLoadUnsignedLong(spec->side);
	ToLoadUnsignedLong(spec->type);
	ToLoadUnsignedLong(spec->icon);
	ToLoadShort(spec->file.vRefNum);
	ToLoadLong(spec->file.parID);
}

void ToSaveMenuSpec(MenuSpec* spec)
{
	ToSaveUnsignedLong(spec->side);
	ToSaveUnsignedLong(spec->type);
	ToSaveUnsignedLong(spec->icon);
	ToSaveShort(spec->file.vRefNum);
	ToSaveLong(spec->file.parID);
}

void ToLoadAppleMenuSettings(AppleMenuSettings* settings)
{
	ToLoadUnsignedLong(settings->color);
	ToLoadShort(settings->showSystemWhen);
	ToLoadShort(settings->modifierKey);
	ToLoadUnsignedLong(settings->mouseWithin);
}

void ToSaveAppleMenuSettings(AppleMenuSettings* settings)
{
	ToSaveUnsignedLong(settings->color);
	ToSaveShort(settings->showSystemWhen);
	ToSaveShort(settings->modifierKey);
	ToSaveUnsignedLong(settings->mouseWithin);
}

void ToLoadGeneralSettings(GeneralSettings* settings)
{
	ToLoadShort(settings->openAtStartup);
	ToLoadShort(settings->finderFrontStartup);
	ToLoadShort(settings->showIconsMainMenu);
	ToLoadShort(settings->showIconsSubmenus);
	ToLoadShort(settings->submenuIconLimit);
}

void ToSaveGeneralSettings(GeneralSettings* settings)
{
	ToSaveShort(settings->openAtStartup);
	ToSaveShort(settings->finderFrontStartup);
	ToSaveShort(settings->showIconsMainMenu);
	ToSaveShort(settings->showIconsSubmenus);
	ToSaveShort(settings->submenuIconLimit);
}

void ToLoadRegistration(Registration* registration)
{
	ToLoadUnsignedLong(registration->code);
	ToLoadUnsignedLong(registration->dateStamp);
}

void ToSaveRegistration(Registration* registration)
{
	ToSaveUnsignedLong(registration->code);
	ToSaveUnsignedLong(registration->dateStamp);
}

Error LoadPreferences()
{
    Error			error;
    FSSpec			file;
    long			length, size;
    OSType			signature;
    short			fileRefNum;
    MenuSpec		menuSpec, *lastMenuSpec;
	AliasHandle		alias;
	FSSpec			menuFileSpec;
	Boolean			wasChanged;

    error=GetPrefsFile(&file);
        TestError(error);
		_i(error)
    
    error=FSpOpenDF(&file, fsRdPerm, &fileRefNum);
    if (error) {
        error=SetDefaultPreferences();
            _g
    }
    
	lastMenuSpec=NULL;
	
    while (1) {
        length=sizeof(signature);
        error=FSRead(fileRefNum, &length, &signature);
		ToLoadUnsignedLong(signature);
            TestError(error);
            _i2(error)
			        
        if (signature==prefsEndMarker)
            break;
        
        length=sizeof(size);
        error=FSRead(fileRefNum, &length, &size);
		ToLoadLong(size);
            TestError(error);
            _i2(error)
			
		// since we don't know how much of these structures are loaded, i.e. if there's a shorter version on disk,
		// it is safest to convert all fields to the file system (big endian) format, then load, then convert back
		
        switch (signature) {
            case prefsRegistration:
				ToSaveRegistration(&registration);
                error=LoadPreference(fileRefNum, size, &registration, sizeof(registration));
				ToLoadRegistration(&registration);
                break;

            case prefsAppleSettings:
				ToSaveAppleMenuSettings(&appleMenuSettings);
				error=LoadPreference(fileRefNum, size, &appleMenuSettings, sizeof(appleMenuSettings));
				ToLoadAppleMenuSettings(&appleMenuSettings);
				break;
			
			case prefsGeneralSettings:
				ToSaveGeneralSettings(&generalSettings);
				error=LoadPreference(fileRefNum, size, &generalSettings, sizeof(generalSettings));
				ToLoadGeneralSettings(&generalSettings);
				break;
			
			case prefsMenuSpec:
                ToSaveMenuSpec(&menuSpec); // probably no point to this since it has no sensible default value, but do anyway
				error=LoadPreference(fileRefNum, size, &menuSpec, sizeof(MenuSpec));
				ToLoadMenuSpec(&menuSpec);
                    _i2(error)
				
                lastMenuSpec=AddMenuToPreferences(&error, menuSpec.side, menuSpec.type, menuSpec.icon, &menuSpec.file);
                break;
			
			case prefsMenuAlias:
				alias=(AliasHandle)NewHandle(size);
				error=TestMemError(alias);
					_i2(error)
					
				HLock((Handle)alias);
				error=LoadPreference(fileRefNum, size, *alias, GetHandleSize((Handle)alias));
				HUnlock((Handle)alias);
					_i3(error)
				
				error=ResolveAlias(NULL, alias, &menuFileSpec, &wasChanged);
					_i3(error)
				
				if (lastMenuSpec)
					lastMenuSpec->file=menuFileSpec;

				_e3
				DisposeHandle((Handle)alias);
				break;
			
            case prefsSelectionCount:
                error=LoadPreference(fileRefNum, size, &selectionCount, sizeof(selectionCount));
				ToLoadUnsignedLong(selectionCount);
                break;
           
            default: // skip over bit of prefs file not understood
                error=SetFPos(fileRefNum, fsFromMark, size);
                    TestError(error);
                break;
        }

        _i2(error)
    }
	
    _e2
    FSClose(fileRefNum);
    _e
    return error;
}

Error LoadPreference(short fileRefNum, long sizeInFile, void* loadPtr, long sizeInMemory)
{
	Error		error;
    long		difference;
	
    if (sizeInFile>sizeInMemory) {
        difference=sizeInFile-sizeInMemory;
		sizeInFile=sizeInMemory;
    } else
        difference=0;

	error=FSRead(fileRefNum, &sizeInFile, loadPtr);
		TestError(error);
        _i(error)
	
    if (difference>0) {
        error=SetFPos(fileRefNum, fsFromMark, difference); // skip over extra bit in file
            TestError(error);
            _i(error)
    }
    
    _e
    return error;
}

Error SavePreferences()
{
    Error		error;
    FSSpec		file;
    short		fileRefNum;
    MenuSpec	*menuSpec;
	AliasHandle	alias;
    
    error=GetPrefsFile(&file);
        _i(error)
    
    FSpDelete(&file); // ignore error
    
    error=FSpCreate(&file, 'ClaM', 'pref', smSystemScript);
        TestError(error);
        _i(error)

    error=FSpOpenDF(&file, fsWrPerm, &fileRefNum);
        TestError(error);
        _i(error)
    
	ToSaveRegistration(&registration);
    error=SavePreference(fileRefNum, prefsRegistration, &registration, sizeof(registration));
	ToLoadRegistration(&registration);
        _i2(error)
	
	ToSaveAppleMenuSettings(&appleMenuSettings);
	error=SavePreference(fileRefNum, prefsAppleSettings, &appleMenuSettings, sizeof(appleMenuSettings));
	ToLoadAppleMenuSettings(&appleMenuSettings);
		_i2(error)
	
    for (menuSpec=firstMenuSpec; menuSpec; menuSpec=menuSpec->next) {
        ToSaveMenuSpec(menuSpec);
		error=SavePreference(fileRefNum, prefsMenuSpec, menuSpec, sizeof(MenuSpec));
		ToLoadMenuSpec(menuSpec);
            _i2(error)
		
		if (menuSpec->type==menuTypeFolder)
			if (NewAlias(NULL, &menuSpec->file, &alias)==noErr) {
				HLock((Handle)alias);
				error=SavePreference(fileRefNum, prefsMenuAlias, *alias, GetHandleSize((Handle)alias));
				HUnlock((Handle)alias);
				DisposeHandle((Handle)alias);
					_i2(error)
			}
    }
    
	ToSaveUnsignedLong(selectionCount);
    error=SavePreference(fileRefNum, prefsSelectionCount, &selectionCount, sizeof(selectionCount));
	ToLoadUnsignedLong(selectionCount);
        _i2(error)
        
	ToSaveGeneralSettings(&generalSettings);
	error=SavePreference(fileRefNum, prefsGeneralSettings, &generalSettings, sizeof(generalSettings));
	ToLoadGeneralSettings(&generalSettings);
		_i2(error)
	
    error=SavePreference(fileRefNum, prefsEndMarker, 0, 0);
        _i2(error)
    
    _e2
    FSClose(fileRefNum);
    _e
    return error;
}

Error SavePreference(short fileRefNum, OSType signature, void* savePtr, long saveSize)
{
	Error	error;
	long	length;
	
	length=sizeof(signature);
	ToSaveUnsignedLong(signature);
	error=FSWrite(fileRefNum, &length, &signature);
	ToLoadUnsignedLong(signature); // not needed but keep for symmetry's sake
		_i(error)
		
	length=sizeof(saveSize);
	ToSaveLong(saveSize);
	error=FSWrite(fileRefNum, &length, &saveSize);
	ToLoadLong(saveSize); // needed for next bit which saves the data
		_i(error)
		
	if (saveSize>0) {
        length=saveSize;
        error=FSWrite(fileRefNum, &length, savePtr);
            _i(error)
    }
	
	_e
	return error;
}

MenuSpec* AddMenuToPreferences(Error* error, OSType side, OSType type, OSType icon, FSSpec* file)
{
    MenuSpec	*menuSpec;
    
    menuSpec=AllocPtr(error, sizeof(MenuSpec));
        _i(*error)
    
    menuSpec->side=side;
    menuSpec->type=type;
    menuSpec->icon=icon;
    
    if (file) // only some need this
        menuSpec->file=*file;
    
    AddToLinkedList((void**)&firstMenuSpec, menuSpec, 0); // add to end of list
    
    _e
    return menuSpec;
}

void RemoveMenuFromPreferences(MenuSpec* menuSpec)
{
    RemoveFromLinkedList((void**)&firstMenuSpec, menuSpec);
    DestroyPtr(menuSpec);
}

Error GetPrefsFile(FSSpec* file)
{
    Error				error;
    const StrFileName	classicMenuPreferences="\pClassic Menu Preferences";
    
    error=FindFolder(kOnAppropriateDisk, kPreferencesFolderType, kCreateFolder, &file->vRefNum, &file->parID);
        TestError(error);
        _i(error)
    
    BlockMove(classicMenuPreferences, file->name, sizeof(StrFileName));
 
     _e
    return error;
}

Error SetDefaultPreferences()
{
    Error				error;
    FSSpec				fileSpec, targetSpec;
	FSRef				targetRef, testRef, parentRef;
    long				folderDirID;
	int					iterations;
    const StrFileName	classicMenuItems="\pClassic Menu Items";
    
    Assert(firstMenuSpec==0);
    
    error=FindFolder(kOnAppropriateDisk, kPreferencesFolderType, kCreateFolder, &fileSpec.vRefNum, &fileSpec.parID);
        TestError(error);
        _i(error)
        
    BlockMove(classicMenuItems, fileSpec.name, sizeof(StrFileName));
    
    FSpDirCreate(&fileSpec, smSystemScript, &folderDirID); // create if doesn't already exist
    
	if (FSFindFolder(kOnAppropriateDisk, kCurrentUserFolderType, kDontCreateFolder, &targetRef)==noErr)
		if (FSGetCatalogInfo(&targetRef, kFSCatInfoNone, NULL, NULL, &targetSpec, NULL)==noErr)
			CreateAlias(&targetSpec, &fileSpec);

	if (FSFindFolder(kLocalDomain, kApplicationsFolderType, kDontCreateFolder, &targetRef)==noErr)
		if (FSGetCatalogInfo(&targetRef, kFSCatInfoNone, NULL, NULL, &targetSpec, NULL)==noErr)
			CreateAlias(&targetSpec, &fileSpec);

	// seems to be best way to get to top of hierarchy
	testRef=targetRef;
	for (iterations=10; iterations>0; iterations--) {
		if (FSGetCatalogInfo(&testRef, kFSCatInfoNone, NULL, NULL, NULL, &parentRef))
			break;
		
		targetRef=testRef;
		testRef=parentRef;
	}
	
	if (FSGetCatalogInfo(&targetRef, kFSCatInfoNone, NULL, NULL, &targetSpec, NULL)==noErr)
		CreateAlias(&targetSpec, &fileSpec);
		
	AddMenuToPreferences(&error, sideLeftMenuBar, menuTypeFolder, menuIconClassicApple, &fileSpec);
        _i(error)
        
    AddMenuToPreferences(&error, sideRightMenuBar, menuTypeProcess, menuIconCurrentApplication, 0);
        _i(error)
    
	ApplyOpenAtStartupSetting();

    _e
    return error;
}

void ApplyOpenAtStartupSetting()
{
	int					items, item;
	char				*itemPath;
	FSRef				fileRef, myFileRef;
	LSItemInfoRecord	info;
	ProcessSerialNumber	process;
	char				appPath[2048];
	Boolean				hasMyFileRefToAdd;
	
	hasMyFileRefToAdd=false;
	
	if (generalSettings.openAtStartup)
		if (GetCurrentProcess(&process)==noErr)
			if (GetProcessBundleLocation(&process, &myFileRef)==noErr)
				hasMyFileRefToAdd=true;

	items=GetCountOfLoginItems(kCurrentUser);
	
	for (item=items-1; item>=0; item--) {
		itemPath=ReturnLoginItemPropertyAtIndex(kCurrentUser, kFullPathInfo, item);

		if (itemPath)
			if (FSPathMakeRef((unsigned char*)itemPath, &fileRef, NULL)==noErr)
				if (LSCopyItemInfoForRef(&fileRef, kLSRequestTypeCreator, &info)==noErr)
					if ((info.filetype=='APPL')&&(info.creator=='ClaM')) {
						if (!hasMyFileRefToAdd)
							RemoveLoginItemAtIndex(kCurrentUser, item);
						else {
							if (FSCompareFSRefs(&fileRef, &myFileRef)==noErr)
								hasMyFileRefToAdd=false;
							else
								RemoveLoginItemAtIndex(kCurrentUser, item);
						}
					}
	}
	
	if (hasMyFileRefToAdd)
		if (FSRefMakePath(&myFileRef, (UInt8*)appPath, sizeof(appPath))==noErr)
			AddLoginItemWithPropertiesToUser(kCurrentUser, appPath, kDoNotHideOnLaunch);
}

pascal Boolean ShowPreferencesFilter(DialogRef dialog, EventRecord* event, short *itemHit);
pascal void DrawSeparatorUserItem(DialogRef theDialog, DialogItemIndex itemNo);

enum {
	openAutoButton=1, finderButton,
	colorText, rainbowPicture, graphitePicture, bluePicture, redPicture, greenPicture,
		rainbowButton, graphiteButton, blueButton, redButton, greenButton,
	systemText, mouseButton, topButton, bottomButton,
		leftButton, rightButton, insideButton, outsideButton,
		keyDownButton, keyUpButton, commandButton, controlButton, shiftButton,
	iconText, mainButton, subButton, subItems, itemsText,
	separator1, separator2, separator3,
	closePseudoItem
};

void ShowPreferencesDialog()
{
	ProcessSerialNumber	process;
	DialogPtr			dialog;
	DialogItemIndex		itemHit;
	ModalFilterUPP		filterUPP;
	UserItemUPP			separatorUPP;
	
	BeginUserInteraction(&process);
	
	dialog=GetNewDialog(130, 0L, (WindowRef)-1L);
	filterUPP=NewModalFilterUPP(ShowPreferencesFilter);
	separatorUPP=NewUserItemUPP(DrawSeparatorUserItem);
	
	SetItemHandle(dialog, separator1, (Handle)separatorUPP);
	SetItemHandle(dialog, separator2, (Handle)separatorUPP);
	SetItemHandle(dialog, separator3, (Handle)separatorUPP);
	
	goto displaySettings;
	
	while (1) {
		ModalDialog(filterUPP, &itemHit);
		
		switch (itemHit) {
			case openAutoButton:
				generalSettings.openAtStartup=ToggleItem(dialog, itemHit);
				ApplyOpenAtStartupSetting();
				goto displaySettings;
			
			case finderButton:
				generalSettings.finderFrontStartup=ToggleItem(dialog, itemHit);
				goto displaySettings;
			
			case rainbowPicture:
			case rainbowButton:
				appleMenuSettings.color=colorRainbow;
				goto changedAppleColor;
			
			case graphitePicture:
			case graphiteButton:
				appleMenuSettings.color=colorGraphite;
				goto changedAppleColor;
				
			case bluePicture:
			case blueButton:
				appleMenuSettings.color=colorBlue;
				goto changedAppleColor;
				
			case redPicture:
			case redButton:
				appleMenuSettings.color=colorRed;
				goto changedAppleColor;
				
			case greenPicture:
			case greenButton:
				appleMenuSettings.color=colorGreen;
				goto changedAppleColor;
				
			case keyDownButton:
				appleMenuSettings.showSystemWhen=showSystemWhenModifierDown;
				goto displaySettings;
			
			case keyUpButton:
				appleMenuSettings.showSystemWhen=showSystemWhenModifierUp;
				goto displaySettings;
			
			case commandButton:
				appleMenuSettings.modifierKey=cmdKey;
				goto displaySettings;
			
/*			case optionButton:
				appleMenuSettings.modifierKey=optionKey;
				goto displaySettings;
*/			
			case controlButton:
				appleMenuSettings.modifierKey=controlKey;
				goto displaySettings;
			
			case shiftButton:
				appleMenuSettings.modifierKey=shiftKey;
				goto displaySettings;
			
			case mouseButton:
				appleMenuSettings.showSystemWhen=showSystemWhenMouseWithin;
				goto displaySettings;
			
			case topButton:
				appleMenuSettings.mouseWithin=mouseWithinTop;
				goto displaySettings;
			
			case bottomButton:
				appleMenuSettings.mouseWithin=mouseWithinBottom;
				goto displaySettings;
			
			case leftButton:
				appleMenuSettings.mouseWithin=mouseWithinLeft;
				goto displaySettings;
			
			case rightButton:
				appleMenuSettings.mouseWithin=mouseWithinRight;
				goto displaySettings;
			
			case insideButton:
				appleMenuSettings.mouseWithin=mouseWithinInside;
				goto displaySettings;
			
			case outsideButton:
				appleMenuSettings.mouseWithin=mouseWithinOutside;
				goto displaySettings;

			case mainButton:
				generalSettings.showIconsMainMenu=ToggleItem(dialog, itemHit);
				DestroyFolderCachedMenu(NULL); // resets all
				goto displaySettings;
			
			case subButton:
				generalSettings.showIconsSubmenus=ToggleItem(dialog, itemHit);
				goto displaySettings;
			
			case subItems:
				generalSettings.submenuIconLimit=GetItemNumber(dialog, itemHit);
				break; // don't display settings 'cause it will be a pain when blank

			changedAppleColor:
				InvalidateReadyMenus();
				goto displaySettings;
			
			displaySettings:
				SetItemValue(dialog, openAutoButton, generalSettings.openAtStartup);
				SetItemValue(dialog, finderButton, generalSettings.finderFrontStartup);
				
				SetItemValue(dialog, rainbowButton, appleMenuSettings.color==colorRainbow);
				SetItemValue(dialog, graphiteButton, appleMenuSettings.color==colorGraphite);
				SetItemValue(dialog, blueButton, appleMenuSettings.color==colorBlue);
				SetItemValue(dialog, redButton, appleMenuSettings.color==colorRed);
				SetItemValue(dialog, greenButton, appleMenuSettings.color==colorGreen);
				
				SetItemValue(dialog, keyDownButton, appleMenuSettings.showSystemWhen==showSystemWhenModifierDown);
				SetItemValue(dialog, keyUpButton, appleMenuSettings.showSystemWhen==showSystemWhenModifierUp);
				SetItemValue(dialog, mouseButton, appleMenuSettings.showSystemWhen==showSystemWhenMouseWithin);
				
				SetItemValue(dialog, commandButton, appleMenuSettings.modifierKey==cmdKey);
/*				SetItemValue(dialog, optionButton, appleMenuSettings.modifierKey==optionKey);
*/				SetItemValue(dialog, controlButton, appleMenuSettings.modifierKey==controlKey);
				SetItemValue(dialog, shiftButton, appleMenuSettings.modifierKey==shiftKey);
				
				SetItemValue(dialog, topButton, appleMenuSettings.mouseWithin==mouseWithinTop);
				SetItemValue(dialog, bottomButton, appleMenuSettings.mouseWithin==mouseWithinBottom);
				SetItemValue(dialog, leftButton, appleMenuSettings.mouseWithin==mouseWithinLeft);
				SetItemValue(dialog, rightButton, appleMenuSettings.mouseWithin==mouseWithinRight);
				SetItemValue(dialog, insideButton, appleMenuSettings.mouseWithin==mouseWithinInside);
				SetItemValue(dialog, outsideButton, appleMenuSettings.mouseWithin==mouseWithinOutside);
				
				SetItemValue(dialog, mainButton, generalSettings.showIconsMainMenu);
				SetItemValue(dialog, subButton, generalSettings.showIconsSubmenus);
				SetItemNumber(dialog, subItems, generalSettings.submenuIconLimit);
				
				SetItemEnabled(dialog, commandButton, (appleMenuSettings.showSystemWhen==showSystemWhenModifierDown)||(appleMenuSettings.showSystemWhen==showSystemWhenModifierUp));
/*				SetItemEnabled(dialog, optionButton, appleMenuSettings.showSystemWhen==showSystemWhenModifierKey);*/
				SetItemEnabled(dialog, controlButton, (appleMenuSettings.showSystemWhen==showSystemWhenModifierDown)||(appleMenuSettings.showSystemWhen==showSystemWhenModifierUp));
				SetItemEnabled(dialog, shiftButton, (appleMenuSettings.showSystemWhen==showSystemWhenModifierDown)||(appleMenuSettings.showSystemWhen==showSystemWhenModifierUp));
				
				SetItemEnabled(dialog, topButton, appleMenuSettings.showSystemWhen==showSystemWhenMouseWithin);
				SetItemEnabled(dialog, bottomButton, appleMenuSettings.showSystemWhen==showSystemWhenMouseWithin);
				SetItemEnabled(dialog, leftButton, appleMenuSettings.showSystemWhen==showSystemWhenMouseWithin);
				SetItemEnabled(dialog, rightButton, appleMenuSettings.showSystemWhen==showSystemWhenMouseWithin);
				SetItemEnabled(dialog, insideButton, appleMenuSettings.showSystemWhen==showSystemWhenMouseWithin);
				SetItemEnabled(dialog, outsideButton, appleMenuSettings.showSystemWhen==showSystemWhenMouseWithin);
				break;
				
			case closePseudoItem:
				goto endDialog;
		}
	}
	
	endDialog:
	DisposeDialog(dialog);
	DisposeModalFilterUPP(filterUPP);
	DisposeUserItemUPP(separatorUPP);
	
	EndUserInteraction(&process);
}

pascal Boolean ShowPreferencesFilter(DialogRef dialog, EventRecord* event, short *itemHit)
{
	WindowRef	window;
	char		character;
	
	if (event->what==mouseDown)
		switch (FindWindow(event->where, &window)) {
			case inDrag:
				*itemHit=0;
				DragWindow(window, event->where, NULL);
				return true;
				
			case inGoAway:
				if (TrackGoAway(window, event->where)) {
					*itemHit=closePseudoItem;
					return true;
				}
				break;
		}
		
	else if ((event->what==keyDown)||(event->what==autoKey)) {
		character=event->message&charCodeMask;
		
		if (
			(character==0x03)||
			(character==0x0D)||
			(character==0x1B)||
			((event->modifiers&cmdKey)&&((character=='W')||(character=='w')))
		) {
			*itemHit=closePseudoItem;
			return true;
		}
	}
					
	return false;
}

pascal void DrawSeparatorUserItem(DialogRef dialog, DialogItemIndex item)
{
	Rect		rect;
	
	GetItemRect(dialog, item, &rect);
	DrawThemeSeparator(&rect, kThemeStateInactive);
}