#include "Essentials.h"
#include "Ready Menus.h"
#include "Filing.h"

static void             RepositionReadyMenus();
static void             RevisibleReadyMenu(ReadyMenu* readyMenu);
static ReadyMenu        *FindReadyMenu(WindowRef window);
static void             DrawReadyMenu(ReadyMenu* readyMenu);
static void             DrawCenteredPNGPicture(CGContextRef cgContext, CFStringRef pictureName, const Rect* rect, float alpha);
static void             SelectReadyMenu(ReadyMenu* readyMenu, DragReference dragRef);

ReadyMenu				*firstReadyMenu=0;
static ReadyMenu		*pulledReadyMenu=0;
static Rect				menuBarRect;

int                     userInteractionHideLevel=0;
Boolean                 classicFrontHideFlag=false, system105OrLater;

WindowGroupRef          menuWindowGroup;
pascal OSStatus         MenuWindowCarbonEvent(EventHandlerCallRef inHandlerRef, EventRef inEvent, void *inUserData);
EventHandlerUPP         MenuWindowCarbonEventUPP=0;

pascal OSStatus         FrontSwitchedCarbonEvent(EventHandlerCallRef inHandlerRef, EventRef inEvent, void *inUserData);
EventHandlerUPP         FrontSwitchedCarbonEventUPP=0;
EventHandlerRef         FrontSwitchedCarbonEventRef;

pascal void             AppleMenuCarbonTimer(EventLoopTimerRef inTimer, void *inUserData);
EventLoopTimerUPP       AppleMenuCarbonTimerUPP=0;
EventLoopTimerRef       AppleMenuCarbonTimerRef;

pascal OSErr			MenuDragTrackHandler(DragTrackingMessage, WindowRef, void*, DragRef);
pascal OSErr			MenuDragReceiveHandler(WindowRef, void*, DragRef);
pascal OSErr			MenuDragInputProc(Point * mouse, SInt16 * modifiers, void * dragInputRefCon, DragRef theDrag);
	
DragTrackingHandlerUPP	MenuDragTrackHandlerUPP=0;
DragReceiveHandlerUPP	MenuDragReceiveHandlerUPP=0;
DragInputUPP			MenuDragInputProcUPP=0;

OSStatus InitialiseReadyMenus()
{
	OSStatus		error;
	EventTypeSpec	frontSwitchedEvents[] = {
						{ kEventClassApplication, kEventAppFrontSwitched },
					};
    long            gestaltResponse;
	
	system105OrLater=(!Gestalt(gestaltSystemVersion, &gestaltResponse)&&(gestaltResponse>=0x1050));
    
    InsertMenu(GetMenu(224), -1);
	InsertMenu(GetMenu(225), -1);

	error=CreateWindowGroup(0, &menuWindowGroup);
		TestError(error);
		_i(error)
	
	error=SetWindowGroupLevel(menuWindowGroup, 31);
		TestError(error);
		_i(error)
	
	MenuWindowCarbonEventUPP=NewEventHandlerUPP(MenuWindowCarbonEvent);
	
	LSInit(kLSInitializeDefaults);
	
	FrontSwitchedCarbonEventUPP=NewEventHandlerUPP(FrontSwitchedCarbonEvent);
	error=InstallEventHandler(GetApplicationEventTarget(), FrontSwitchedCarbonEventUPP,
			sizeof(frontSwitchedEvents)/sizeof(EventTypeSpec), frontSwitchedEvents, 0,
			&FrontSwitchedCarbonEventRef);
		TestError(error);
		_i(error)
	
	AppleMenuCarbonTimerUPP=NewEventLoopTimerUPP(AppleMenuCarbonTimer);
	error=InstallEventLoopTimer(GetMainEventLoop(), kEventDurationNoWait, kEventDurationSecond/15,
			AppleMenuCarbonTimerUPP, 0, &AppleMenuCarbonTimerRef);
		TestError(error);
		_i(error)
	
//	HideMenuBar(); // this used to be necessary otherwise our menu overlay window doesn't work
//	but it causes problems in Panther (OS X 10.3) making the menu not appear immediately. Weird.

	ScreenInformationChanged();
 
	_e
	return error;
}

void ScreenInformationChanged()
{
	RepositionReadyMenus();
}

Error RebuildReadyMenus()
{
	Error		error;
	MenuSpec	*menuSpec;
	
	ClearReadyMenus();
	
	error=0; // in case no menu specs
	for (menuSpec=firstMenuSpec; menuSpec; menuSpec=menuSpec->next) {
		if (menuSpec->side==sideLeftMenuBar) { // for this version which is left side only
			AddReadyMenu(&error, menuSpec);
				_i(error)
		}
	}
	
	_e
	return error;
}

void RevisibleReadyMenus()
{
	ReadyMenu	*readyMenu;
	
	for (readyMenu=firstReadyMenu; readyMenu; readyMenu=readyMenu->next)
		RevisibleReadyMenu(readyMenu);
}

void RevisibleReadyMenu(ReadyMenu* readyMenu)
{
	ShowHide(readyMenu->window, /*(userInteractionHideLevel<=0)&&*/(!classicFrontHideFlag)&&(!readyMenu->appleMenuInvisibleFlag));
    DrawReadyMenu(readyMenu); // seems needed with new overlay window class
}

void InvalidateReadyMenus()
{
	ReadyMenu	*readyMenu;
	Rect		portRect;

	for (readyMenu=firstReadyMenu; readyMenu; readyMenu=readyMenu->next)
		InvalWindowRect(readyMenu->window, GetWindowPortBounds(readyMenu->window, &portRect));
}

void ClearReadyMenus()
{
	while (firstReadyMenu)
		RemoveReadyMenu(firstReadyMenu);
}

void DestroyReadyMenus()
{
	Assert(firstReadyMenu==0);
	
	RemoveEventLoopTimer(AppleMenuCarbonTimerRef);
	DisposeEventLoopTimerUPP(AppleMenuCarbonTimerUPP);
	RemoveEventHandler(FrontSwitchedCarbonEventRef);
	DisposeEventHandlerUPP(FrontSwitchedCarbonEventUPP);
	LSTerm();
	DisposeEventHandlerUPP(MenuWindowCarbonEventUPP);
	ReleaseWindowGroup(menuWindowGroup);
}

pascal OSStatus FrontSwitchedCarbonEvent(EventHandlerCallRef inHandlerRef, EventRef inEvent, void *inUserData)
{
	ProcessSerialNumber		process;
	ProcessInfoRec			infoRec;
	FSSpec					fileSpec;
	FSRef					fileRef;
	LSItemInfoRecord		itemInfo;
	
	GetFrontProcess(&process);
	
	infoRec.processInfoLength=sizeof(infoRec);
	infoRec.processName=NULL;
	infoRec.processAppSpec=&fileSpec;
	
	classicFrontHideFlag=false;
	
    if (GetProcessBundleLocation(&process, &fileRef)==0)
        if (LSCopyItemInfoForRef(&fileRef, kLSRequestAllFlags, &itemInfo)==0) {
            if (
                ((itemInfo.flags&kLSItemInfoIsClassicApp)&&!(itemInfo.flags&kLSItemInfoIsNativeApp))
                ||((itemInfo.flags&kLSItemInfoAppPrefersClassic)&&(itemInfo.flags&kLSItemInfoIsNativeApp))
            )
                classicFrontHideFlag=true;
        }
					
	RevisibleReadyMenus();
	
	return 0;
}

void SelectReadyMenu(ReadyMenu* readyMenu, DragReference dragRef)
{
	Point			clickPoint;
 
	if (pulledReadyMenu==0) {
		pulledReadyMenu=readyMenu;
		DrawReadyMenu(readyMenu);
		QDFlushPortBuffer(GetWindowPort(readyMenu->window), NULL);
		
		clickPoint.h=readyMenu->rect.left;
		clickPoint.v=readyMenu->rect.bottom+4;
		
		switch (readyMenu->spec->type) {
			case menuTypeFolder:
				ClickFolderMenu(clickPoint, readyMenu, dragRef);
				break;
		}
		
		pulledReadyMenu=0;
		DrawReadyMenu(readyMenu);
	}
}

pascal OSStatus MenuWindowCarbonEvent(EventHandlerCallRef inHandlerRef, EventRef inEvent, void *inUserData)
{
	WindowRef			window;
	EventParamType		gotType;
	UInt32				gotSize;
	ReadyMenu			*readyMenu;
	ProcessSerialNumber myProcess, process;
	HIPoint				clickLoc;
    Point               qdClickLoc;
	GrafPtr				oldPort;
   
    switch (GetEventClass(inEvent)) {
        case kEventClassMouse:
            switch (GetEventKind(inEvent)) {
                case kEventMouseDown:
                    if (GetEventParameter(inEvent, kEventParamWindowRef, typeWindowRef,
                            &gotType, sizeof(window), &gotSize, &window)==0)
                        readyMenu=FindReadyMenu(window);
                    else
                        readyMenu=0;
                  
                    if (readyMenu&&(GetEventParameter(inEvent, kEventParamMouseLocation, typeHIPoint,
                            &gotType, sizeof(clickLoc), &gotSize, &clickLoc)==0)) { // all for special test
                        
                        qdClickLoc.h=clickLoc.x;
                        qdClickLoc.v=clickLoc.y;
                        
                        GetPort(&oldPort);
                        SetPortWindowPort(readyMenu->window);
                        LocalToGlobal(&qdClickLoc);
                        SetPort(oldPort);

                        if ((qdClickLoc.h==1)&&(qdClickLoc.v==18))
                            if (FindApplicationProcess('KM*E', &process)==0)
                                break;
                    }
                    
                    if (userInteractionHideLevel>0) {
                        GetCurrentProcess(&myProcess);
                        SetFrontProcess(&myProcess);
                    } else
                        SelectReadyMenu(readyMenu, NULL);
                break;
            }
            break;
            
        case kEventClassWindow:
            if (GetEventParameter(inEvent, kEventParamDirectObject, typeWindowRef,
                    &gotType, sizeof(window), &gotSize, &window)==0)
                readyMenu=FindReadyMenu(window);
            else
                readyMenu=0;
            
            switch (GetEventKind(inEvent)) {
                case kEventWindowConstrain:
                    break;
               
                case kEventWindowShown:
                case kEventWindowDrawContent:
                    DrawReadyMenu(readyMenu);
                    break;
            }
            break;
    }
	
	return noErr;
}

pascal void AppleMenuCarbonTimer(EventLoopTimerRef inTimer, void *inUserData)
{
	ReadyMenu			*readyMenu;
	Point				globalPoint, localPoint;
	Rect				windowRect, localRect;
	Boolean				makeInvisible;
	UInt32				modifiers;
	GrafPtr				oldPort;

	if (pulledReadyMenu) {

	} else if (Button()) {
		// do nothing (could be accessing a normal menu bar menu)
	
	} else {
		GetGlobalMouse(&globalPoint);

		for (readyMenu=firstReadyMenu; readyMenu; readyMenu=readyMenu->next) {
			makeInvisible=0;
		
			GetWindowBounds(readyMenu->window, kWindowContentRgn, &windowRect);
			
			if (readyMenu->spec->side==sideLeftMenuBar) {
				if (PtInRect(globalPoint, &windowRect)) {
					if ((appleMenuSettings.showSystemWhen==showSystemWhenModifierDown)||(appleMenuSettings.showSystemWhen==showSystemWhenModifierUp)) {
						modifiers=GetCurrentKeyModifiers();
						if (((short)modifiers)&appleMenuSettings.modifierKey)
							makeInvisible=1;
					
					} else if (appleMenuSettings.showSystemWhen==showSystemWhenMouseWithin) {
						localPoint=globalPoint;
						GetPort(&oldPort);
						SetPortWindowPort(readyMenu->window);
						GlobalToLocal(&localPoint);
						SetPort(oldPort);
						
						switch (appleMenuSettings.mouseWithin) {
							case mouseWithinLeft:
								if (localPoint.h<((readyMenu->rect.left+readyMenu->rect.right)/2))
									makeInvisible=1;
								break;
							
							case mouseWithinRight:
								if (localPoint.h>=((readyMenu->rect.left+readyMenu->rect.right)/2))
									makeInvisible=1;
								break;
								
							case mouseWithinTop:
								if (localPoint.v<((readyMenu->rect.top+readyMenu->rect.bottom)/2))
									makeInvisible=1;
								break;
								
							case mouseWithinBottom:
								if (localPoint.v>=((readyMenu->rect.top+readyMenu->rect.bottom)/2))
									makeInvisible=1;
								break;
								
							case mouseWithinInside:
								localRect=readyMenu->rect;
								InsetRect(&localRect, (localRect.right-localRect.left)/4, (localRect.bottom-localRect.top)/4);
								if (PtInRect(localPoint, &localRect))
									makeInvisible=1;
								break;
								
							case mouseWithinOutside:
								localRect=readyMenu->rect;
								InsetRect(&localRect, (localRect.right-localRect.left)/4, (localRect.bottom-localRect.top)/4);
								if (!PtInRect(localPoint, &localRect))
									makeInvisible=1;
								break;
						}
					}
				}
				
				if (appleMenuSettings.showSystemWhen==showSystemWhenModifierUp)
					makeInvisible=!makeInvisible;
			}
		
			if (makeInvisible!=readyMenu->appleMenuInvisibleFlag) {
				readyMenu->appleMenuInvisibleFlag=makeInvisible;
				RevisibleReadyMenu(readyMenu);
			}

		}
	}
}

ReadyMenu* AddReadyMenu(Error* error, MenuSpec* menuSpec)
{
	ReadyMenu		*readyMenu;
	Rect			rect;
	EventTypeSpec	windowEvents[] = {
						{ kEventClassWindow, kEventWindowConstrain },
						{ kEventClassWindow, kEventWindowDrawContent },
                        { kEventClassWindow, kEventWindowShown },
						{ kEventClassMouse, kEventMouseDown }
					};
	 
	readyMenu=AllocPtr(error, sizeof(ReadyMenu));
		_i(*error)
	
	readyMenu->spec=menuSpec;
	readyMenu->appleMenuInvisibleFlag=0;
	
	SetRect(&rect, 0, 0, 100, 100);
	
	*error=CreateNewWindow(kOverlayWindowClass,
			kWindowNoShadowAttribute|kWindowStandardHandlerAttribute|kWindowOpaqueForEventsAttribute|kWindowHideOnFullScreenAttribute, &rect, &readyMenu->window);
		TestError(*error);
		_i(*error)
		
	SetWindowGroup(readyMenu->window, menuWindowGroup);
    if (system105OrLater)
		HIWindowChangeAvailability(readyMenu->window, 1<<8 /* kHIWindowCanJoinAllSpaces */, 0); // only in API from 10.4 onwards
 
	*error=InstallEventHandler(GetWindowEventTarget(readyMenu->window), MenuWindowCarbonEventUPP,
			sizeof(windowEvents)/sizeof(EventTypeSpec), windowEvents, 0, NULL);
		TestError(*error);
		_i(*error)
	
	if (!MenuDragTrackHandlerUPP)
		MenuDragTrackHandlerUPP=NewDragTrackingHandlerUPP(MenuDragTrackHandler);
	
	if (MenuDragTrackHandlerUPP)
		InstallTrackingHandler(MenuDragTrackHandlerUPP, readyMenu->window, readyMenu);
	
	if (!MenuDragReceiveHandlerUPP)
		MenuDragReceiveHandlerUPP=NewDragReceiveHandlerUPP(MenuDragReceiveHandler);
	
	if (MenuDragReceiveHandlerUPP)
		InstallReceiveHandler(MenuDragReceiveHandler, readyMenu->window, readyMenu);
	
	readyMenu->menu=NULL;
	
    // commented out because on first run on OS X 10.5, this can pre-cache folder contents while new items still being added
    // must be something to do with new asynchronous file writing or some such. FlushVol didn't help.
    
	/*switch (readyMenu->spec->type) {
        case menuTypeFolder:
			InitializeFolderCachedMenu(readyMenu);
			break;
	}*/
	
	AddToLinkedList((void**)&firstReadyMenu, readyMenu, 0); // add to end of list (tho' it doesn't matter)
	
	RepositionReadyMenus();
   
	ShowWindow(readyMenu->window);

	DrawReadyMenu(readyMenu);

	_e
	return readyMenu;
}

void RemoveReadyMenu(ReadyMenu* readyMenu)
{
	DisposeWindow(readyMenu->window);
	
	switch (readyMenu->spec->type) {
		case menuTypeFolder:
			DestroyFolderCachedMenu(readyMenu);
			break;
	}
	
	RemoveFromLinkedList((void**)&firstReadyMenu, readyMenu);
	DestroyPtr(readyMenu);

	RepositionReadyMenus();
}

void RepositionReadyMenus()
{
	Rect		remainingBarRect, windowRect;
	ReadyMenu	*readyMenu;
	short		menuBarHeight;
	
	menuBarHeight=22; // don't query OS since it returns wrong result
	menuBarRect=(*GetMainDevice())->gdRect;
	menuBarRect.bottom=menuBarRect.top+menuBarHeight;
	 
	remainingBarRect=menuBarRect;
	for (readyMenu=firstReadyMenu; readyMenu; readyMenu=readyMenu->next) {
		windowRect=remainingBarRect;
					
		switch (readyMenu->spec->side) {
			case sideLeftMenuBar:
				windowRect.right=windowRect.left+45;
				remainingBarRect.left=windowRect.right;
				break;
				
			case sideRightMenuBar:
				windowRect.left=windowRect.right-menuBarHeight;
				remainingBarRect.right=windowRect.left;
				break;
		}
		
		SetWindowBounds(readyMenu->window, kWindowContentRgn, &windowRect);
		GetWindowPortBounds(readyMenu->window, &readyMenu->rect);
		
		if (readyMenu->spec->side==sideLeftMenuBar)
			readyMenu->rect.left+=10;
	}
}

ReadyMenu* FindReadyMenu(WindowRef window)
{
	ReadyMenu	*readyMenu;
	
	for (readyMenu=firstReadyMenu; readyMenu; readyMenu=readyMenu->next)
		if (readyMenu->window==window)
			break;
			
	return readyMenu;
}

void DrawReadyMenu(ReadyMenu* readyMenu)
{
	Rect                        windowRect;
	CFStringRef                 pictureName;
	CGContextRef                cgContext;
	CGRect                      cgRect, cgSelectRect;
    HIThemeMenuTitleDrawInfo    drawInfo;
	
	pictureName=NULL; 
	
	switch (readyMenu->spec->icon) {
		case menuIconClassicApple:
			switch (appleMenuSettings.color) {
				case colorBlue:
					pictureName=CFSTR("blue.png");
					break;
					
				case colorRed:
					pictureName=CFSTR("red.png");
					break;
					
				case colorGreen:
					pictureName=CFSTR("green.png");
					break;
					
				case colorGraphite:
					pictureName=CFSTR("grey.png");
					break;
					
				case colorRainbow:
					pictureName=CFSTR("rainbow.png");
					break;
			}
			break;
	}
	
    if (QDBeginCGContext(GetWindowPort(readyMenu->window), &cgContext)==noErr) {
        GetWindowPortBounds(readyMenu->window, &windowRect);
        cgRect=CGRectMake(0, 0, windowRect.right-windowRect.left, windowRect.bottom-windowRect.top);
        CGContextClearRect(cgContext, cgRect);
        
        if ((readyMenu==pulledReadyMenu) || !system105OrLater) {
            drawInfo.version=0;
            drawInfo.state=(readyMenu==pulledReadyMenu) ? kThemeMenuSelected : kThemeMenuActive;
            drawInfo.attributes=0;
            drawInfo.condensedTitleExtra=0;
            cgSelectRect=CGRectMake(readyMenu->rect.left, readyMenu->rect.top,
                readyMenu->rect.right-readyMenu->rect.left, readyMenu->rect.bottom-readyMenu->rect.top);
            
            CGContextSaveGState(cgContext);
            HIThemeDrawMenuTitle(&cgRect, &cgSelectRect, &drawInfo, cgContext, kHIThemeOrientationNormal, NULL);
            CGContextRestoreGState(cgContext);
        }
        
        if (pulledReadyMenu==readyMenu)
            DrawCenteredPNGPicture(cgContext, CFSTR("white.png"), &readyMenu->rect, 1.0);

        DrawCenteredPNGPicture(cgContext, pictureName, &readyMenu->rect, system105OrLater ? 0.5 : 0.8);
 
        QDEndCGContext(GetWindowPort(readyMenu->window), &cgContext);
    }
}

void DrawCenteredPNGPicture(CGContextRef cgContext, CFStringRef pictureName, const Rect* rect, float alpha)
{
	CFURLRef			cfUrl;
	CGRect				cgRect;
	CGDataProviderRef	cgDataProvider;
    CGImageRef          cgImage;

    if (pictureName) {
        cfUrl=CFBundleCopyResourceURL(CFBundleGetMainBundle(), pictureName, NULL, NULL);
        if (cfUrl) {
            cgDataProvider=CGDataProviderCreateWithURL(cfUrl);
            
            if (cgDataProvider) {
                cgImage=CGImageCreateWithPNGDataProvider(cgDataProvider, NULL, false, kCGRenderingIntentDefault);
                
                if (cgImage) {
                    cgRect=CGRectMake((rect->right+rect->left-CGImageGetWidth(cgImage))/2,
                        (rect->bottom+rect->top-CGImageGetHeight(cgImage))/2,
                        CGImageGetWidth(cgImage), CGImageGetHeight(cgImage));
                    
                    CGContextSaveGState(cgContext);
                    CGContextSetAlpha(cgContext, alpha);
                    CGContextDrawImage(cgContext, cgRect, cgImage);
                    CGContextRestoreGState(cgContext);
                    
                    CGImageRelease(cgImage);
                }
                
                CGDataProviderRelease(cgDataProvider);
            }
            
            CFRelease(cfUrl);
        }
    }
}

void ChangeReadyMenuFrontProcessIcon()
{
	ReadyMenu	*readyMenu;
	Rect		windowRect;
	
	for (readyMenu=firstReadyMenu; readyMenu; readyMenu=readyMenu->next)
		if (readyMenu->spec->icon==menuIconCurrentApplication)
			InvalWindowRect(readyMenu->window, GetWindowPortBounds(readyMenu->window, &windowRect));
}

void BeginUserInteraction(ProcessSerialNumber* priorProcess)
{
	ProcessSerialNumber myProcess;
	
	userInteractionHideLevel++;
	RevisibleReadyMenus();
	
	GetFrontProcess(priorProcess);
	GetCurrentProcess(&myProcess);
	SetFrontProcess(&myProcess);
}

void EndUserInteraction(const ProcessSerialNumber* priorProcess)
{
	if (priorProcess)
		SetFrontProcess(priorProcess);

	userInteractionHideLevel--;
	RevisibleReadyMenus();
}

static Boolean CheckDragAcceptable(ReadyMenu* menu, DragRef dragRef);
static OSErr AcceptDrag(ReadyMenu* menu, DragRef dragRef);

pascal OSErr MenuDragTrackHandler(DragTrackingMessage message, WindowRef window, void* refCon, DragRef dragRef)
{
	Rect			rect;
	RgnHandle		region;
	Boolean			showHilite;
	
	showHilite=false;
	if (message==kDragTrackingInWindow)
		showHilite=CheckDragAcceptable((ReadyMenu*)refCon, dragRef);
		
	if (showHilite) {
		region=NewRgn();
		GetWindowPortBounds(window, &rect);
		RectRgn(region, &rect);
		ShowDragHilite(dragRef, region, true);
		DisposeRgn(region);

	} else
		HideDragHilite(dragRef);	

	return 0;
}

pascal OSErr MenuDragReceiveHandler(WindowRef window, void* refCon, DragRef dragRef)
{
	OSErr		error;
	
	if (CheckDragAcceptable((ReadyMenu*)refCon, dragRef))
		error=AcceptDrag((ReadyMenu*)refCon, dragRef);
	else
		error=dragNotAcceptedErr;
	
	return error;
}

Boolean CheckDragAcceptable(ReadyMenu* menu, DragRef dragRef)
{
	OSErr			error;
	unsigned short	items, item, flavor, flavors;
	DragItemRef		itemRef;
	FlavorType		type;

	if (menu->spec->type==menuTypeFolder) {
		error=CountDragItems(dragRef, &items);
			_i(error)
			
		for (item=1; item<=items; item++) {
			error=GetDragItemReferenceNumber(dragRef, item, &itemRef);
				_i(error)
				
			error=CountDragItemFlavors(dragRef, itemRef, &flavors);
				_i(error)
				
			for (flavor=1; flavor<=flavors; flavor++) {
				error=GetFlavorType(dragRef, itemRef, flavor, &type);
					_i(error)
					
				if (type==kDragFlavorTypeHFS)
					goto itemOK;
			}
			
			_g
	
			itemOK:;
		}
		
		return true;
	}
	
	_e
	return false;
}

OSErr AcceptDrag(ReadyMenu* menu, DragRef dragRef)
{
	OSErr				error;
	unsigned short		items, item;
	DragItemRef			itemRef;
	Size				size;
	HFSFlavor			hfsFlavor;
	
	error=CountDragItems(dragRef, &items);
		_i(error)
		
	for (item=1; item<=items; item++) {
		error=GetDragItemReferenceNumber(dragRef, item, &itemRef);
			_i(error)
			
		size=sizeof(hfsFlavor);
		error=GetFlavorData(dragRef, itemRef, kDragFlavorTypeHFS, &hfsFlavor, &size, 0);
			_i(error)
		
		if (menu->spec->type==menuTypeFolder) {
			error=CreateAlias(&hfsFlavor.fileSpec, &menu->spec->file);
				_i(error)
				
			DestroyFolderCachedMenu(menu);
		}
	}
	
	_e
	return error;
}