#include "Essentials.h"
#include "Filing.h"
#include "Ready Menus.h"

OSStatus	InstallCarbonEventHandlers();
void		RemoveCarbonEventHandlers();

OSStatus	InstallAppleEventHandlers();
void		RemoveAppleEventHandlers();

int main(int argc, char *argv[])
{
    OSStatus				error;
    
    InitCursor();

    InstallStandardEventHandler(GetApplicationEventTarget());
    
    error=InstallAppleEventHandlers();
        _i(error)
    
    error=InstallCarbonEventHandlers();
        _i2(error)
    
    error=InitialiseReadyMenus();
        _i3(error)
        
    error=LoadPreferences();
        _i4(error)
    
	error=RebuildReadyMenus();
        _i4(error)
    
	if (generalSettings.finderFrontStartup)
		BringFinderToFront();
	
	RunApplicationEventLoop();
    
    SavePreferences();
    ClearReadyMenus();
    _e4
    DestroyReadyMenus();
    _e3
    RemoveCarbonEventHandlers();
    _e2
    RemoveAppleEventHandlers();
    _e
    return 0;
}
 
EventHandlerUPP		VolumeCarbonEventUPP=0;
pascal OSStatus		VolumeCarbonEvent(EventHandlerCallRef inHandlerRef,
                        EventRef inEvent, void *inUserData);
EventHandlerRef		VolumeCarbonEventRef;

OSStatus InstallCarbonEventHandlers()
{
    OSStatus		error;
    EventTypeSpec	volumeEvents[] = {
                        { kEventClassVolume, kEventVolumeMounted },
                        { kEventClassVolume, kEventVolumeUnmounted },
                    };

    VolumeCarbonEventUPP=NewEventHandlerUPP(VolumeCarbonEvent);

    error=InstallEventHandler(GetApplicationEventTarget(), VolumeCarbonEventUPP,
            sizeof(volumeEvents)/sizeof(EventTypeSpec), volumeEvents, 0, &VolumeCarbonEventRef);
			
	TestError(error);
	
	return error;
}

void RemoveCarbonEventHandlers()
{
    RemoveEventHandler(VolumeCarbonEventRef);
    DisposeEventHandlerUPP(VolumeCarbonEventUPP);
}

pascal OSStatus VolumeCarbonEvent(EventHandlerCallRef inHandlerRef, EventRef inEvent, void *inUserData)
{
    switch (GetEventKind(inEvent)) {
        case kEventVolumeMounted:
		case kEventVolumeUnmounted:
            DestroyFolderCachedMenu(NULL); // resets all
            break;
    }
    
    return noErr;
}

pascal OSErr		AppleEventQuitApp(const AppleEvent*, AppleEvent*, SInt32);
AEEventHandlerUPP	AppleEventQuitAppUPP;

OSStatus InstallAppleEventHandlers()
{
    OSStatus	error;
    
    AppleEventQuitAppUPP=NewAEEventHandlerUPP(AppleEventQuitApp);
    
	error=AEInstallEventHandler(kCoreEventClass, kAEQuitApplication, AppleEventQuitAppUPP, 0, false);
    
	TestError(error);
	
    return error;
}

void RemoveAppleEventHandlers()
{
    AERemoveEventHandler(kCoreEventClass, kAEQuitApplication, AppleEventQuitAppUPP, false);
    DisposeAEEventHandlerUPP(AppleEventQuitAppUPP);
}

pascal OSErr AppleEventQuitApp(const AppleEvent* eventIn, AppleEvent* eventOut, SInt32 refNum)
{
    QuitApplicationEventLoop(); // this is required even though we install standard event handler
    
    return 0;
}
